@@ -21,9 +21,24 @@ return;
 }
 
 
-void pfatal(char *err) {
-  perror(err);
-  fprintf(stderr, "\n");
-  exit(1);
+
+void pfatal(char *err, ...) {
+va_list  ap;va_start(ap, err);
+fflush(stdout);
+vfprintf(stderr, err, ap);
+va_end(ap);
+perror(" ");
+fflush(stderr);
+exit(1);
 }
 
+
+void gh_perror(char *err, ...) {
+va_list  ap;va_start(ap, err);
+fflush(stdout);
+vfprintf(stderr, err, ap);
+va_end(ap);
+perror(" ");
+fflush(stderr);
+return;
+}
