@@ -15,7 +15,7 @@ int res = 0;
 va_start(ap, format);
 
 if (!warning) {
-  error("WARNING:  your system apparrently does not offer snprintf().  Reverting to less secure version");
+  error("WARNING:  your system apparently does not offer snprintf().  Reverting to less secure version");
   warning = 1;
 }
 
@@ -36,7 +36,9 @@ len = strlen(buf);
 if (len >= SNPRINTF_BUFSIZE)
   fatal("Our bufferZ may have been overfl0wed!!@#$!@#");
 
-strcpy(str, buf);
+
+Strncpy(str, buf, n);
 va_end (ap);
+if (len >= n) return -1;
 return len;
 }
