@@ -1,10 +1,11 @@
 #include "tcpip.h"
 
-
 /* Globals */
 int jumpok = 0;
 static jmp_buf jmp_env;
 
+extern struct ops o;
+
 /* Sig_ALRM handler */
 void sig_alarm(int signo) {
 if (jumpok)
@@ -42,11 +43,25 @@ char dev[128];
   return 0;
 }
 
+/* Calls pcap_open_live and spits out an error (and quits) if the call faile.
+   So a valid pcap_t will always be returned. */
+pcap_t *my_pcap_open_live(char *device, int snaplen, int promisc, int to_ms) 
+{
+  char err0r[PCAP_ERRBUF_SIZE];
+  pcap_t *pt;
+  if (!((pt = pcap_open_live(device, snaplen, promisc, to_ms, err0r)))) {
+    fatal("pcap_open_live: %s\nThere are several possible reasons for this, depending on your operating system:\n"
+          "LINUX: If you are getting Socket type not supported, try modprobe af_packet or recompile your kernel with SOCK_PACKET enabled.\n"
+          "*BSD:  If you are getting device not configured, you need to recompile your kernel with Berkeley Packet Filter support.  If you are getting No such file or directory, try creating the device (eg cd /dev; MAKEDEV <device>; or use mknod).\n"
+          "SOLARIS:  If you are trying to scan localhost and getting '/dev/lo0: No such file or directory', complain to Sun.  I don't think Solaris can support advanced localhost scans.  You can probably use \"-P0 -sT localhost\" though.\n\n", err0r);
+  }
+  return pt;
+}
 
 /* Standard swiped internet checksum routine */
 inline unsigned short in_cksum(unsigned short *ptr,int nbytes) {
 
-register long           sum;            /* assumes long == 32 bits */
+register int            sum;            /* XXX assumes long == 32 bits */
 u_short                 oddbyte;
 register u_short        answer;         /* assumes u_short == 16 bits */
 
@@ -100,18 +115,35 @@ int resolve(char *hostname, struct in_addr *ip) {
   return 0;
 }
 
+int send_tcp_raw_decoys( int sd, struct in_addr *victim, unsigned short sport, 
+			 unsigned short dport, unsigned int seq,
+			 unsigned int ack, unsigned char flags,
+			 unsigned short window, char *options, int optlen,
+			 char *data, unsigned short datalen) 
+{
+  int decoy;
+
+  for(decoy = 0; decoy < o.numdecoys; decoy++) 
+    if (send_tcp_raw(sd, &o.decoys[decoy], victim, sport, dport, seq, ack,
+		     flags, window, options, optlen, data, datalen) == -1)
+      return -1;
+
+  return 0;
+}
+
+
 int send_tcp_raw( int sd, struct in_addr *source, 
 		  struct in_addr *victim, unsigned short sport, 
-		  unsigned short dport, unsigned long seq,
-		  unsigned long ack, unsigned char flags,
+		  unsigned short dport, unsigned int seq,
+		  unsigned int ack, unsigned char flags,
 		  unsigned short window, char *options, int optlen,
 		  char *data, unsigned short datalen) 
 {
 
 struct pseudo_header { 
   /*for computing TCP checksum, see TCP/IP Illustrated p. 145 */
-  unsigned long s_addy;
-  unsigned long d_addr;
+  unsigned int s_addy;
+  unsigned int d_addr;
   char zer0;
   unsigned char protocol;
   unsigned short length;
@@ -230,8 +262,8 @@ ip->ip_sum = in_cksum((unsigned short *)ip, sizeof(struct ip));
 #endif
 
 if (TCPIP_DEBUGGING > 1) {
-printf("Raw TCP packet creation completed!  Here it is:\n");
-readtcppacket(packet,BSDUFIX(ip->ip_len));
+  log_write(LOG_STDOUT, "Raw TCP packet creation completed!  Here it is:\n");
+  readtcppacket(packet,BSDUFIX(ip->ip_len));
 }
 
 res = Sendto("send_tcp_raw", sd, packet, BSDUFIX(ip->ip_len), 0,
@@ -252,7 +284,7 @@ int sleeptime = 0;
 
 do {
   if (TCPIP_DEBUGGING > 1) {  
-    printf("trying sendto(%d, packet, %d, 0, %s, %d)",
+    log_write(LOG_STDOUT, "trying sendto(%d, packet, %d, 0, %s, %d)",
 	   sd, len, inet_ntoa(sin->sin_addr), tolen);
   }
   if ((res = sendto(sd, packet, len, flags, to, tolen)) == -1) {
@@ -270,7 +302,7 @@ do {
 } while( res == -1);
 
 if (TCPIP_DEBUGGING > 1)
-  printf("successfully sent %d bytes of raw_tcp!\n", res);
+  log_write(LOG_STDOUT, "successfully sent %d bytes of raw_tcp!\n", res);
 
 return res;
 }
@@ -318,10 +350,10 @@ if (ip->ip_p== IPPROTO_TCP) {
 
     printf("ttl: %hu ", ip->ip_ttl);
 
-    if (tcp->th_flags & (TH_SYN | TH_ACK)) printf("Seq: %lu\tAck: %lu\n", 
-						  (unsigned long) ntohl(tcp->th_seq), (unsigned long) ntohl(tcp->th_ack));
-    else if (tcp->th_flags & TH_SYN) printf("Seq: %lu\n", (unsigned long) ntohl(tcp->th_seq));
-    else if (tcp->th_flags & TH_ACK) printf("Ack: %lu\n", (unsigned long) ntohl(tcp->th_ack));
+    if (tcp->th_flags & (TH_SYN | TH_ACK)) printf("Seq: %u\tAck: %u\n", 
+						  (unsigned int) ntohl(tcp->th_seq), (unsigned int) ntohl(tcp->th_ack));
+    else if (tcp->th_flags & TH_SYN) printf("Seq: %u\n", (unsigned int) ntohl(tcp->th_seq));
+    else if (tcp->th_flags & TH_ACK) printf("Ack: %u\n", (unsigned int) ntohl(tcp->th_ack));
   }
 }
 if (readdata && i < tot_len) {
@@ -375,6 +407,19 @@ if (ip->ip_p== IPPROTO_UDP) {
  return 0;
 }
 
+int send_udp_raw_decoys( int sd, struct in_addr *victim, unsigned short sport, 
+		  unsigned short dport, char *data, unsigned short datalen) {
+  int decoy;
+  
+  for(decoy = 0; decoy < o.numdecoys; decoy++) 
+    if (send_udp_raw(sd, &o.decoys[decoy], victim, sport, dport, data, 
+		     datalen) == -1)
+      return -1;
+
+  return 0;
+}
+
+
 
 int send_udp_raw( int sd, struct in_addr *source, 
 		  struct in_addr *victim, unsigned short sport, 
@@ -488,6 +533,165 @@ free(packet);
 return res;
 }
 
+int send_small_fragz_decoys(int sd, struct in_addr *victim, unsigned long seq, 
+			    int sport, int dport, int flags) {
+  int decoy;
+
+  for(decoy = 0; decoy < o.numdecoys; decoy++) 
+    if (send_small_fragz(sd, &o.decoys[decoy], victim, seq, sport, 
+				dport, 
+				flags) == -1)
+      return -1;
+
+  return 0;
+}
+
+/* Much of this is swiped from my send_tcp_raw function above, which 
+   doesn't support fragmentation */
+int send_small_fragz(int sd, struct in_addr *source, struct in_addr *victim,
+		     unsigned long seq, int sport, int dport, int flags)
+ {
+
+struct pseudo_header { 
+/*for computing TCP checksum, see TCP/IP Illustrated p. 145 */
+  unsigned long s_addy;
+  unsigned long d_addr;
+  char zer0;
+  unsigned char protocol;
+  unsigned short length;
+};
+/*In this placement we get data and some field alignment so we aren't wasting
+  too much to compute the TCP checksum.*/
+
+char packet[sizeof(struct ip) + sizeof(struct tcphdr) + 100];
+struct ip *ip = (struct ip *) packet;
+struct tcphdr *tcp = (struct tcphdr *) (packet + sizeof(struct ip));
+struct pseudo_header *pseudo = (struct pseudo_header *) (packet + sizeof(struct ip) - sizeof(struct pseudo_header)); 
+char *frag2 = packet + sizeof(struct ip) + 16;
+struct ip *ip2 = (struct ip *) (frag2 - sizeof(struct ip));
+static int myttl = 0;
+int res;
+struct sockaddr_in sock;
+int id;
+
+if (!myttl)  myttl = (time(NULL) % 14) + 51;
+
+/* It was a tough decision whether to do this here for every packet
+   or let the calling function deal with it.  In the end I grudgingly decided
+   to do it here and potentially waste a couple microseconds... */
+sethdrinclude(sd);
+
+
+/*Why do we have to fill out this damn thing? This is a raw packet, after all */
+sock.sin_family = AF_INET;
+sock.sin_port = htons(dport);
+
+sock.sin_addr.s_addr = victim->s_addr;
+
+bzero((char *)packet, sizeof(struct ip) + sizeof(struct tcphdr));
+
+pseudo->s_addy = source->s_addr;
+pseudo->d_addr = victim->s_addr;
+pseudo->protocol = IPPROTO_TCP;
+pseudo->length = htons(sizeof(struct tcphdr));
+
+tcp->th_sport = htons(sport);
+tcp->th_dport = htons(dport);
+tcp->th_seq = (seq)? htonl(seq) : get_random_uint();
+
+tcp->th_off = 5 /*words*/;
+tcp->th_flags = flags;
+
+tcp->th_win = htons(2048); /* Who cares */
+
+tcp->th_sum = in_cksum((unsigned short *)pseudo, 
+		       sizeof(struct tcphdr) + sizeof(struct pseudo_header));
+
+/* Now for the ip header of frag1 */
+
+bzero((char *) packet, sizeof(struct ip)); 
+ip->ip_v = 4;
+ip->ip_hl = 5;
+/*RFC 791 allows 8 octet frags, but I get "operation not permitted" (EPERM)
+  when I try that.  */
+ip->ip_len = BSDFIX(sizeof(struct ip) + 16);
+id = ip->ip_id = get_random_uint();
+ip->ip_off = BSDFIX(MORE_FRAGMENTS);
+ip->ip_ttl = myttl;
+ip->ip_p = IPPROTO_TCP;
+ip->ip_src.s_addr = source->s_addr;
+ip->ip_dst.s_addr = victim->s_addr;
+
+#if HAVE_IP_IP_SUM
+ip->ip_sum= in_cksum((unsigned short *)ip, sizeof(struct ip));
+#endif
+if (o.debugging > 1) {
+  log_write(LOG_STDOUT, "Raw TCP packet fragment #1 creation completed!  Here it is:\n");
+  hdump(packet,20);
+}
+if (o.debugging > 1) 
+  log_write(LOG_STDOUT, "\nTrying sendto(%d , packet, %d, 0 , %s , %d)\n",
+	 sd, ntohs(ip->ip_len), inet_ntoa(*victim),
+	 (int) sizeof(struct sockaddr_in));
+/* Lets save this and send it AFTER we send the second one, just to be
+   cute ;) */
+
+if ((res = sendto(sd, packet,sizeof(struct ip) + 16 , 0, 
+		  (struct sockaddr *)&sock, sizeof(struct sockaddr_in))) == -1)
+  {
+    perror("sendto in send_syn_fragz");
+    return -1;
+  }
+if (o.debugging > 1) log_write(LOG_STDOUT, "successfully sent %d bytes of raw_tcp!\n", res);
+
+/* Create the second fragment */
+
+bzero((char *) ip2, sizeof(struct ip));
+ip2->ip_v= 4;
+ip2->ip_hl = 5;
+ip2->ip_len = BSDFIX(sizeof(struct ip) + 4); /* the rest of our TCP packet */
+ip2->ip_id = id;
+ip2->ip_off = BSDFIX(2);
+ip2->ip_ttl = myttl;
+ip2->ip_p = IPPROTO_TCP;
+ip2->ip_src.s_addr = source->s_addr;
+ip2->ip_dst.s_addr = victim->s_addr;
+
+#if HAVE_IP_IP_SUM
+ip2->ip_sum = in_cksum((unsigned short *)ip2, sizeof(struct ip));
+#endif
+if (o.debugging > 1) {
+  log_write(LOG_STDOUT, "Raw TCP packet fragment creation completed!  Here it is:\n");
+  hdump(packet,20);
+}
+if (o.debugging > 1) 
+
+  log_write(LOG_STDOUT, "\nTrying sendto(%d , ip2, %d, 0 , %s , %d)\n", sd, 
+	 ntohs(ip2->ip_len), inet_ntoa(*victim), (int) sizeof(struct sockaddr_in));
+if ((res = sendto(sd, (void *)ip2,sizeof(struct ip) + 4 , 0, 
+		  (struct sockaddr *)&sock, (int) sizeof(struct sockaddr_in))) == -1)
+  {
+    perror("sendto in send_tcp_raw frag #2");
+    return -1;
+  }
+
+return 1;
+}
+
+int send_ip_raw_decoys( int sd, struct in_addr *victim, unsigned char proto,
+			char *data, unsigned short datalen) {
+
+  int decoy;
+
+  for(decoy = 0; decoy < o.numdecoys; decoy++) 
+    if (send_ip_raw(sd, &o.decoys[decoy], victim, proto, data, 
+			   datalen) == -1)
+      return -1;
+
+  return 0;
+
+
+}
 
 int send_ip_raw( int sd, struct in_addr *source, 
 		  struct in_addr *victim, unsigned char proto,
@@ -748,108 +952,92 @@ exit(1);
 
 /* Read an IP packet using libpcap .  We return the packet and take
    a pcap descripter and a pointer to the packet length (which we set
-   in the function.  If you want a read timeout, specify one in 
-   pcap_open_live(). If you want a maximum length returned, you also
+   in the function. If you want a maximum length returned, you
    should specify that in pcap_open_live() */
 
-char *readip_pcap(pcap_t *pd, unsigned int *len) {
-static int offset = -1;
-static pcap_t *lastpcap = NULL;
+/* to_usec is the timeout period in microseconds -- use 0 to skip the
+   test and -1 to block forever.  Note that we don't interrupt pcap, so
+   low values (and 0) degenerate to the timeout specified 
+   in pcap_open_live()
+ */
+
+char *readip_pcap(pcap_t *pd, unsigned int *len, long to_usec) {
+int offset = -1;
 struct pcap_pkthdr head;
 char *p;
 int datalink;
+int timedout = 0;
+struct timeval tv_start, tv_end;
 
 if (!pd) fatal("NULL packet device passed to readip_pcap");
-if (!lastpcap || pd != lastpcap) { 
-  /* New packet capture device, need to recompute offset */
-  if ( (datalink = pcap_datalink(pd)) < 0)
-    fatal("Cannot obtain datalink information: %s", pcap_geterr(pd));
-  switch(datalink) {
-  case DLT_EN10MB: offset = 14; break;
-  case DLT_IEEE802: offset = 22; break;
-  case DLT_NULL: offset = 4; break;
-  case DLT_SLIP:
-#if (FREEBSD || OPENBSD || NETBSD || BSDI)
-    offset = 16;
-#else
-    offset = 24; /* Anyone use this??? */
+
+/* New packet capture device, need to recompute offset */
+ if ( (datalink = pcap_datalink(pd)) < 0)
+   fatal("Cannot obtain datalink information: %s", pcap_geterr(pd));
+ switch(datalink) {
+ case DLT_EN10MB: offset = 14; break;
+ case DLT_IEEE802: offset = 22; break;
+#ifdef DLT_LOOP
+ case DLT_LOOP:
+#endif
+ case DLT_NULL: offset = 4; break;
+ case DLT_SLIP:
+#ifdef DLT_SLIP_BSDOS
+ case DLT_SLIP_BSDOS:
 #endif
-    break;
-  case DLT_PPP: 
 #if (FREEBSD || OPENBSD || NETBSD || BSDI)
-    offset = 4;
+   offset = 16;
+#else
+   offset = 24; /* Anyone use this??? */
+#endif
+   break;
+ case DLT_PPP: 
+#ifdef DLT_PPP_BSDOS
+ case DLT_PPP_BSDOS:
+#endif
+#if (FREEBSD || OPENBSD || NETBSD || BSDI)
+   offset = 4;
 #else
 #ifdef SOLARIS
-    offset = 8;
+   offset = 8;
 #else
-    offset = 24; /* Anyone use this? */
+   offset = 24; /* Anyone use this? */
 #endif /* ifdef solaris */
 #endif /* if freebsd || openbsd || netbsd || bsdi */
-    break;
-  case DLT_RAW: offset = 0; break;
-  default: fatal("Unknown datalink type (%d)", datalink);
-  }
-}
-lastpcap = pd;
-do {
-  p = (char *) pcap_next(pd, &head);
-  if (p)
-    p += offset;
-  else {
-    /* timed out */ 
-    *len=0;
-    return NULL;
-  }
-} while(!p || (*p & 0x40) != 0x40); /* Go until we get IPv4 packet */
-*len = head.caplen - offset;
-return p;
-}
+   break;
+ case DLT_RAW: offset = 0; break;
+ case DLT_FDDI: offset = 21; break;
+#ifdef DLT_ENC
+ case DLT_ENC: offset = 12; break;
+#endif
+ default: fatal("Unknown datalink type (%d)", datalink);
+ }
 
-/* Like readip_pcap except we use our own timeout value.  This is needed
-   due to a "bug" in libpcap.  The Linux pcap_open_live takes a timeout
-   but DOES NOT EVEN LOOK AT IT! */
-char *readip_pcap_timed(pcap_t *pd, unsigned int *len, unsigned long timeout /*seconds
- */) {
-static int offset = -1;
-static pcap_t *lastpcap = NULL;
-struct pcap_pkthdr head;
-char *p;
-int datalink;
-
-if (!pd) fatal("NULL packet device passed to readip_pcap");
-if (!lastpcap || pd != lastpcap) {
-  /* New packet capture device, need to recompute offset */
-  if ( (datalink = pcap_datalink(pd)) < 0)
-    fatal("Cannot obtain datalink information: %s", pcap_geterr(pd));
-  switch(datalink) {
-  case DLT_EN10MB: offset = 14; break;
-  case DLT_IEEE802: offset = 22; break;
-  case DLT_NULL: offset = 4; break;
-  case DLT_SLIP:
-  case DLT_PPP: offset = 24; break;
-  case DLT_RAW: offset = 0; break;
-  default: fatal("Unknown datalink type (%d)", datalink);
-  }
-}
-lastpcap = pd;
-signal(SIGALRM, sig_alarm);
-if (setjmp(jmp_env)) {
-  /* We've timed out */
-  *len = 0;
-  return NULL;
-}
-jumpok = 1;
-alarm(timeout);
-do {
-p = (char *) pcap_next(pd, &head);
-if (p)
-  p += offset;
-} while(!p || (*p & 0x40) != 0x40); /* Go until we get IPv4 packet */
-alarm(0);
-jumpok = 0;
-signal(SIGALRM, SIG_DFL);
-*len = head.caplen - offset;
-return p;
+ if (to_usec > 0) {
+   gettimeofday(&tv_start, NULL);
+ }
+ do {
+   p = (char *) pcap_next(pd, &head);
+   if (p)
+     p += offset;
+   if (!p || (*p & 0x40) != 0x40) {
+     /* Should we timeout? */
+     if (to_usec == 0) {
+       timedout = 1;
+     } else if (to_usec > 0) {
+       gettimeofday(&tv_end, NULL);
+       if (TIMEVAL_SUBTRACT(tv_end, tv_start) >= to_usec) {
+	 timedout = 1;     
+       }
+     }
+   }
+ } while(!timedout && (!p || (*p & 0x40) != 0x40)); /* Go until we get IPv4 packet */
+ if (timedout) {
+   *len = 0;
+   return NULL;
+ }
+ *len = head.caplen - offset;
+ return p;
 }
 
 int ipaddr2devname( char *dev, struct in_addr *addr ) {
@@ -916,12 +1104,31 @@ struct interface_info *getinterfaces(int *howmany) {
       fatal("getinterfaces: SIOCGIFCONF claims you have no network interfaces!\n");
 #if HAVE_SOCKADDR_SA_LEN
     /*    len = MAX(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);*/
-    len = ifr->ifr_addr.sa_len;
+    len = ifr->ifr_addr.sa_len + sizeof(ifr->ifr_name);
 #else
-    len = sizeof(SA);
+    len = sizeof(struct ifreq);
+    /* len = sizeof(SA); */
 #endif
+
+#if TCPIP_DEBUGGING
+    printf("ifnet list length = %d\n",ifc.ifc_len);
+    printf("sa_len = %d\n",len);
+    hdump(buf, ifc.ifc_len);
+    printf("ifr = %X\n",(unsigned int)(*(char **)&ifr));
+    printf("Size of struct ifreq: %d\n", sizeof(struct ifreq));
+#endif
+
     for(; ifr && *((char *)ifr) && ((char *)ifr) < buf + ifc.ifc_len; 
-	((*(char **)&ifr) +=  sizeof(ifr->ifr_name) + len )) {
+	((*(char **)&ifr) += len )) {
+#if TCPIP_DEBUGGING
+      printf("ifr_name size = %d\n", sizeof(ifr->ifr_name));
+      printf("ifr = %X\n",(unsigned int)(*(char **)&ifr));
+#endif
+
+      /* skip any device with no name */
+      if (!*((char *)ifr))
+        continue;
+
       sin = (struct sockaddr_in *) &ifr->ifr_addr;
       memcpy(&(mydevs[numinterfaces].addr), (char *) &(sin->sin_addr), sizeof(struct in_addr));
       /* In case it is a stinkin' alias */
@@ -929,6 +1136,12 @@ struct interface_info *getinterfaces(int *howmany) {
 	*p = '\0';
       strncpy(mydevs[numinterfaces].name, ifr->ifr_name, 63);
       mydevs[numinterfaces].name[63] = '\0';
+
+
+#if TCPIP_DEBUGGING
+      printf("Interface %d is %s\n",numinterfaces,mydevs[numinterfaces].name);
+#endif
+
       numinterfaces++;
       if (numinterfaces == 47)  {      
 	error("My god!  You seem to have WAY too many interfaces!  Things may not work right\n");
@@ -936,7 +1149,7 @@ struct interface_info *getinterfaces(int *howmany) {
       }
 #if HAVE_SOCKADDR_SA_LEN
       /* len = MAX(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);*/
-      len = ifr->ifr_addr.sa_len;
+      len = ifr->ifr_addr.sa_len + sizeof(ifr->ifr_name);
 #endif 
       mydevs[numinterfaces].name[0] = '\0';
     }
@@ -953,6 +1166,7 @@ struct interface_info *getinterfaces(int *howmany) {
    source parameter.   Some of the stuff is
    from Stevens' Unix Network Programming V2.  He had an easier suggestion
    for doing this (in the book), but it isn't portable :( */
+#define ROUTETHROUGH_MAXROUTES 512
 char *routethrough(struct in_addr *dest, struct in_addr *source) {
   static int initialized = 0;
   int i;
@@ -962,9 +1176,9 @@ char *routethrough(struct in_addr *dest, struct in_addr *source) {
   struct interface_info *mydevs;
   static struct myroute {
     struct interface_info *dev;
-    unsigned long mask;
-    unsigned long dest;
-  } myroutes[128];
+    unsigned int mask;
+    unsigned int dest;
+  } myroutes[ROUTETHROUGH_MAXROUTES];
   int numinterfaces = 0;
   char *p, *endptr;
   char iface[64];
@@ -992,13 +1206,16 @@ char *routethrough(struct in_addr *dest, struct in_addr *source) {
 	  error("Could not find interface in /proc/net/route line");
 	  continue;
 	}
+	if (*p == '*')
+	  continue; /* Deleted route -- any other valid reason for
+		       a route to start with an asterict? */
 	Strncpy(iface, p, sizeof(iface));
 	if ((p = strchr(iface, ':'))) {
 	  *p = '\0'; /* To support IP aliasing */
 	}
 	p = strtok(NULL, " \t\n");
 	endptr = NULL;
-	myroutes[numroutes].dest = strtol(p, &endptr, 16);
+	myroutes[numroutes].dest = strtoul(p, &endptr, 16);
 	if (!endptr || *endptr) {
 	  error("Failed to determine Destination from /proc/net/route");
 	  continue;
@@ -1012,7 +1229,7 @@ char *routethrough(struct in_addr *dest, struct in_addr *source) {
 	  continue;
 	}
 	endptr = NULL;
-	myroutes[numroutes].mask = strtol(p, &endptr, 16);
+	myroutes[numroutes].mask = strtoul(p, &endptr, 16);
 	if (!endptr || *endptr) {
 	  error("Failed to determine mask from /proc/net/route");
 	  continue;
@@ -1020,7 +1237,7 @@ char *routethrough(struct in_addr *dest, struct in_addr *source) {
 
 
 #if TCPIP_DEBUGGING
-	  printf("#%d: for dev %s, The dest is %lX and the mask is %lX\n", numroutes, iface, myroutes[numroutes].dest, myroutes[numroutes].mask);
+	  printf("#%d: for dev %s, The dest is %X and the mask is %X\n", numroutes, iface, myroutes[numroutes].dest, myroutes[numroutes].mask);
 #endif
 	  for(i=0; i < numinterfaces; i++)
 	    if (!strcmp(iface, mydevs[i].name)) {
@@ -1030,7 +1247,7 @@ char *routethrough(struct in_addr *dest, struct in_addr *source) {
 	  if (i == numinterfaces) 
 	    fatal("Failed to find interface %s mentioned in /proc/net/route\n", iface);
 	  numroutes++;
-	  if (numroutes == 128)
+	  if (numroutes == ROUTETHROUGH_MAXROUTES)
 	    fatal("My god!  You seem to have WAY to many routes!\n");
       }
       fclose(routez);
